from __future__ import absolute_import
from django.conf.urls import *
from django.views.static import serve
import os

JAVASCRIPT_PATH = "%s/js" % os.path.dirname(__file__)

urlpatterns = patterns('ajax.views',
    (r'^(?P<application>\w+)/(?P<model>\w+).json', 'endpoint_loader'),
    (r'^(?P<application>\w+)/(?P<model>\w+)/(?P<method>\w+).json', 'endpoint_loader'),
    (r'^(?P<application>\w+)/(?P<model>\w+)/(?P<pk>\d+)/(?P<method>\w+)/?(?P<taggit_command>(add|remove|set|clear|similar))?.json$', 'endpoint_loader'),
    (r'^js/(?P<path>.*)$', serve,
        {'document_root': JAVASCRIPT_PATH}),
)



from __future__ import absolute_import
import django

if django.VERSION >= (1, 7):
    from django.utils.module_loading import import_string as path_to_import
    from importlib import import_module
    from logging import getLogger
else:
    # 1.4 LTS compatibility
    from ajax.utils import import_by_path as path_to_import
    from django.utils.importlib import import_module
    from django.utils.log import getLogger



from __future__ import absolute_import
import django.dispatch

ajax_created = django.dispatch.Signal(providing_args=['instance'])
ajax_deleted = django.dispatch.Signal(providing_args=['instance'])
ajax_updated = django.dispatch.Signal(providing_args=['instance'])



from __future__ import absolute_import
import sys

from django.core.exceptions import ImproperlyConfigured
from ajax.compat import import_module


def import_by_path(dotted_path, error_prefix=''):
    """
    Import a dotted module path and return the attribute/class designated by
    the last name in the path. Raise ImproperlyConfigured if something goes
    wrong. This has come straight from Django 1.6
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        raise ImproperlyConfigured("%s%s doesn't look like a module path" % (
            error_prefix, dotted_path))
    try:
        module = import_module(module_path)
    except ImportError as e:
        raise ImproperlyConfigured('%sError importing module %s: "%s"' % (
            error_prefix, module_path, e))
    try:
        attr = getattr(module, class_name)
    except AttributeError:
        raise ImproperlyConfigured(
            '%sModule "%s" does not define a "%s" attribute/class' % (
                error_prefix, module_path, class_name
            )
        )
    return attr



from __future__ import absolute_import
from django.core import serializers
from ajax.exceptions import AlreadyRegistered, NotRegistered
from django.db.models.fields import FieldDoesNotExist
from django.db import models
from django.conf import settings
from django.utils.html import escape
from django.db.models.query import QuerySet
from django.utils.encoding import smart_str
import collections
import six


# Used to change the field name for the Model's pk.
AJAX_PK_ATTR_NAME = getattr(settings, 'AJAX_PK_ATTR_NAME', 'pk')


def _fields_from_model(model):
    return [field.name for field in model.__class__._meta.fields]


class DefaultEncoder(object):
    _mapping = {
        'IntegerField': int,
        'PositiveIntegerField': int,
        'AutoField': int,
        'FloatField': float,
    }

    def to_dict(self, record, expand=False, html_escape=False, fields=None):
        self.html_escape = html_escape
        if hasattr(record, '__exclude__') and callable(record.__exclude__):
            try:
                exclude = record.__exclude__()
                if fields is None:
                    fields = _fields_from_model(record)
                fields = set(fields) - set(exclude)
            except TypeError:
                pass
        data = serializers.serialize('python', [record], fields=fields)[0]

        if hasattr(record, 'extra_fields'):
            ret = record.extra_fields
        else:
            ret = {}

        ret.update(data['fields'])
        ret[AJAX_PK_ATTR_NAME] = data['pk']

        for field, val in six.iteritems(ret):
            try:
                f = record.__class__._meta.get_field(field)
                if expand and isinstance(f, models.ForeignKey):
                    try:
                        row = f.rel.to.objects.get(pk=val)
                        new_value = self.to_dict(row, False)
                    except f.rel.to.DoesNotExist:
                        new_value = None  # Changed this to None from {} -G
                else:
                    new_value = self._encode_value(f, val)

                ret[smart_str(field)] = new_value
            except FieldDoesNotExist as e:
                pass  # Assume extra fields are already safe.
                  
        if expand and hasattr(record, 'tags') and \
          record.tags.__class__.__name__.endswith('TaggableManager'):
          # Looks like this model is using taggit.
          ret['tags'] = [{'name': self._escape(t.name), 
          'slug': self._escape(t.slug)} for t in record.tags.all()]
          
        return ret

    __call__ = to_dict

    def _encode_value(self, field, value):
        if value is None:
            return value # Leave all None's as-is as they encode fine.

        try:
            return self._mapping[field.__class__.__name__](value)
        except KeyError:
            if isinstance(field, models.ForeignKey):
                f = field.rel.to._meta.get_field(field.rel.field_name)
                return self._encode_value(f, value)
            elif isinstance(field, models.BooleanField):
                # If someone could explain to me why the fuck the Python
                # serializer appears to serialize BooleanField to a string
                # with "True" or "False" in it, please let me know.
                return (value == "True" or (type(value) == bool and value))

        return self._escape(value)

    def _escape(self, value):
        if self.html_escape:
            return escape(value)
        return value


class HTMLEscapeEncoder(DefaultEncoder):
    """Encodes all values using Django's HTML escape function."""
    def _escape(self, value):
        return escape(value)


class ExcludeEncoder(DefaultEncoder):
    def __init__(self, exclude):
        self.exclude = exclude

    def __call__(self, record, html_escape=False):
        fields = set(_fields_from_model(record)) - set(self.exclude)
        return self.to_dict(record, html_escape=html_escape, fields=fields)


class IncludeEncoder(DefaultEncoder):
    def __init__(self, include):
        self.include = include

    def __call__(self, record, html_escape=False):
        return self.to_dict(record, html_escape=html_escape, fields=self.include)


class Encoders(object):
    def __init__(self):
        self._registry = {}

    def register(self, model, encoder):
        if model in self._registry:
            raise AlreadyRegistered()

        self._registry[model] = encoder

    def unregister(self, model):
        if model not in self._registry:
            raise NotRegistered()

        del self._registry[model]
    
    def get_encoder_from_record(self, record):
        if isinstance(record, models.Model) and \
            record.__class__ in self._registry:
            encoder = self._registry[record.__class__]
        else:
            encoder = DefaultEncoder()
        return encoder
        
    def encode(self, record, encoder=None, html_escape=False):
        if isinstance(record, collections.Iterable):
            ret = []
            for i in record:
                if not encoder:
                    encoder = self.get_encoder_from_record(i)
                ret.append(self.encode(i, html_escape=html_escape))
        else:
            if not encoder:
                encoder = self.get_encoder_from_record(record)
            ret = encoder(record, html_escape=html_escape)

        return ret


encoder = Encoders()



from __future__ import absolute_import
import json

from django.utils.encoding import smart_str
from django.http import HttpResponse, HttpResponseNotFound, \
    HttpResponseForbidden, HttpResponseNotAllowed, HttpResponseServerError, \
    HttpResponseBadRequest


class AlreadyRegistered(Exception):
    pass


class NotRegistered(Exception):
    pass


class PrimaryKeyMissing(Exception):
    pass


class AJAXError(Exception):
    RESPONSES = {
        400: HttpResponseBadRequest,
        403: HttpResponseForbidden,
        404: HttpResponseNotFound,
        405: HttpResponseNotAllowed,
        500: HttpResponseServerError,
    }

    def __init__(self, code, msg, **kwargs):
        self.code = code
        self.msg = msg
        self.extra = kwargs  # Any kwargs will be appended to the output.

    def get_response(self):
        try:
            msg = smart_str(self.msg.decode())
        except (AttributeError,):
            msg = smart_str(self.msg)
        error = {
            'success': False,
            'data': {
                'code': self.code,
                'message': msg
            }
        }
        error.update(self.extra)

        response = self.RESPONSES[self.code]()
        response.content = json.dumps(error, separators=(',', ':'))
        return response



from __future__ import absolute_import
from __future__ import print_function
import json

import mock

from django.test import TestCase
from django.contrib.auth.models import User
from django.utils import six

from ajax.endpoints import ModelEndpoint
from ajax.exceptions import AJAXError
from ajax.signals import ajax_deleted

from .models import Widget, Category
from .endpoints import WidgetEndpoint, CategoryEndpoint


class BaseTest(TestCase):
    fixtures = ['users.json', 'categories.json', 'widgets.json']

    def setUp(self):
        self.login('jstump')

    def login(self, username, password='testing'):
        user = User.objects.get(username=username)
        login_successful = self.client.login(username=user.username,
            password=password)
        self.assertTrue(login_successful)

    def post(self, uri, data={}, debug=False, status_code=200):
        """Send an AJAX request.

        This handles sending the AJAX request via the built-in Django test
        client and then decodes the response.

        ``status_code`` lets you define what you expect the status code
        to be which will be tested before returning the response object
        and the decoded JSON content.

        ``debug`` if set to True will spit out the response and content.
        """
        response = self.client.post(uri, data)
        if debug:
            print(response.__class__.__name__)
            print(response)

        self.assertEquals(status_code, response.status_code)
        if isinstance(response.content, six.text_type):
            return response, json.loads(response.content)
        else:
            return response, json.loads(response.content.decode('utf-8'))


class EncodeTests(BaseTest):
    def test_encode(self):
        from ajax.encoders import encoder
        widget = Widget.objects.get(pk=1)
        self.assertEquals(widget.title,'Iorem lipsum color bit amit')
        encoded = encoder.encode(widget)
        for k in ('title','active','description'):
            self.assertEquals(encoded[k],getattr(widget,k))
        widgets = Widget.objects.all()
        all_encoded = encoder.encode(widgets)
        for encoded in all_encoded:
            widget = Widget.objects.get(pk=encoded['pk'])
            for k in ('title','active','description'):
                self.assertEquals(encoded[k],getattr(widget,k))


class EndpointTests(BaseTest):
    def test_echo(self):
        """Test the ad-hoc echo endpoint."""
        resp, content = self.post('/ajax/example/echo.json',
            {'name': 'Joe Stump', 'age': 31})
        self.assertEquals('Joe Stump', content['data']['name'])
        self.assertEquals('31', content['data']['age'])

    def test_empty_foreign_key(self):
        """Test that nullable ForeignKey fields can be set to null"""
        resp, content = self.post('/ajax/example/widget/3/update.json',
            {'category': ''})
        self.assertEquals(None, content['data']['category'])
        self.assertEquals(None, Widget.objects.get(pk=3).category)

    def test_false_foreign_key(self):
        """Test that nullable ForeignKey fields can be set to null by setting it to false"""
        resp, content = self.post('/ajax/example/widget/6/update.json',
            {'category': False})
        self.assertEquals(None, content['data']['category'])
        self.assertEquals(None, Widget.objects.get(pk=6).category)

    def test_logged_out_user_fails(self):
        """Make sure @login_required rejects requests to echo."""
        self.client.logout()
        resp, content = self.post('/ajax/example/echo.json', {},
            status_code=403)

    def test_has_changes_does_save(self):
        """Test that updating to a new value calls save."""
        with mock.patch.object(ModelEndpoint, '_save') as mock_save:
            resp, content = self.post('/ajax/example/widget/6/update.json',
                {'active': False})
            self.assertTrue(mock_save.called)

    def test_no_changes_doesnt_save(self):
        """Test that updating to an existing value doesnt call save."""
        with mock.patch.object(ModelEndpoint, '_save') as mock_save:
            resp, content = self.post('/ajax/example/widget/6/update.json',
                {'active': True})
            self.assertFalse(mock_save.called)


class MockRequest(object):
    def __init__(self, **kwargs):
        self.POST = kwargs
        self.user = None


class ModelEndpointTests(BaseTest):
    def setUp(self):
        self.list_endpoint = WidgetEndpoint('example', Widget, 'list')
        self.category_endpoint = CategoryEndpoint('example', Category, 'list')

    def test_list_returns_all_items(self):
        results = self.list_endpoint.list(MockRequest())
        self.assertEqual(len(results.data), Widget.objects.count())

    def test_list_obeys_endpoint_pagination_amount(self):
        self.list_endpoint.max_per_page = 1
        results = self.list_endpoint.list(MockRequest())
        self.assertEqual(len(results.data), 1)

    def test_list__ajaxerror_if_can_list_isnt_set(self):
        self.assertRaises(AJAXError, self.category_endpoint.list, MockRequest())

    def test_out_of_range_returns_empty_list(self):
        results = self.list_endpoint.list(MockRequest(current_page=99))
        self.assertEqual(len(results.data), 0)

    def test_request_doesnt_override_max_per_page(self):
        self.list_endpoint.max_per_page = 1
        results = self.list_endpoint.list(MockRequest(items_per_page=2))
        self.assertEqual(len(results.data), 1)

    def test_list_has_permission__default_empty(self):
        Category.objects.create(title='test')

        self.category_endpoint.can_list = lambda *args, **kwargs: True

        results = self.category_endpoint.list(MockRequest())
        self.assertEqual(0, len(results.data))

    def test_list_has_total(self):
        self.category_endpoint.can_list = lambda *args, **kwargs: True

        results = self.list_endpoint.list(MockRequest())
        self.assertEqual(6, results.metadata['total'])


class ModelEndpointPostTests(TestCase):
    """
    Integration test for full urls->views->endpoint->encoder (and back) cycle.
    """
    def setUp(self):
        for title in ['first', 'second', 'third']:
            Widget.objects.create(title=title, active=True)
        u = User(email='test@example.org', username='test')
        u.set_password('password')
        u.save()

    def test_can_request_list_with_total(self):
        self.client.login(username='test', password='password')

        resp = self.client.post('/ajax/example/widget/list.json')
        if isinstance(resp.content, six.text_type):
            content = json.loads(resp.content)
        else:
            content = json.loads(resp.content.decode('utf-8'))
        self.assertTrue('total' in list(content.keys()))
        self.assertEquals(content['total'], 3)

    def test_delete(self):
        widget = Widget.objects.all()[0]

        # spec attr required for mocking signal in Django 1.4.
        mocked_ajax_delete_signal = mock.Mock(spec=lambda: None)
        ajax_deleted.connect(mocked_ajax_delete_signal)

        self.client.login(username='test', password='password')

        self.client.post('/ajax/example/widget/%s/delete.json' % widget.pk)

        self.assertTrue(mocked_ajax_delete_signal.called)
        self.assertEqual(widget.pk, mocked_ajax_delete_signal.call_args[1]['payload']['pk'])
